/*************************************************************************************************************************/
/* Description:
/* According to the clustering result, a minimum spanning tree (MST) is generated as branch cuts for each residue cluster.

/* Release date of FLSv1.0: Aug.05, 2018
/* Latest updated date of FLSv2.0: Oct.02, 2018
/*************************************************************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <setjmp.h>
#include "brcut.h"
#include "mst.h"
#include "util.h"
#include "file.h"

#define OUTPUTMST	0

// place a branch cut between point (is, js) and point (ie, je)
void PlaceCut(unsigned char *bitflags, int is, int js, int ie, int je, const int xsize, const int ysize) {

	int i, j, m, n, istep, jstep;
	double k;

	if (ie > is) is++;
	else if (ie < is) ie++;
	if (je > js) js++;
	else if (je < js) je++;

	if (ie==is && je==js) {
		bitflags[is*ysize+js] |= CUT;
		return;
	}

	m = (ie>is) ? ie-is : is-ie;
	n = (je>js) ? je-js : js-je;
	if (m > n) {
		istep = (ie>is) ? 1 : -1;
		k = ((double)(je-js)/(double)(ie-is));
		for (i = is; i != ie+istep; i += istep) {
			j = js + k*(i-is) +0.5;
			bitflags[i*ysize+j] |= CUT;
		}
	} else {
		jstep = (je>js) ? 1 : -1;
		k = ((double)(ie-is)/(double)(je-js));
		for (j = js; j != je+jstep; j += jstep) {
			i = is + k*(j-js) + 0.5;
			bitflags[i*ysize+j] |= CUT;
		}
	}
}

// generate MST for all residue cluster
int GenMSTforAllCluster(int *mark_global, unsigned char *bitflags, short *row, short *col, int *type, float *dist, 
						 const int len, const int xsize, const int ysize, const int cluster_limit) {

	int i, j, k, status;
	MGraph *G;
	int bad_flag = 0;
#if OUTPUTMST
	FILE *fp;
#endif

	// mark all residues that will be on the MST
	for (i = 0; i < xsize; ++i) {
		for (j = 0; j < ysize; ++j) {
			k = i*ysize+j;
			bitflags[k] &= ~NEED_TODO;
			if (mark_global[k] != 0) {
				bitflags[k] |= NEED_TODO;
			}
		}
	}

	// traversing the full image to generate all MSTs
	G = (MGraph *)malloc(sizeof(MGraph));
	for (i = 0; i < xsize; ++i) {
		for (j = 0; j < ysize; ++j) {
			k = i*ysize+j;
			if (bitflags[k] & NEED_TODO) {

				status = CreateGraph(G, row, col, type, dist, len, xsize, ysize, cluster_limit, mark_global[k]);		// generate undirected graph
				if (status == 1) {
					fprintf(STDOUT, "Generate MST for cluster: %d...\t", mark_global[k]);
					MinSpanTreePrim(G, k, bitflags, xsize, ysize);					// get MST and place branch cuts
					DestroyGraph(G);												//
					fprintf(STDOUT, "Done.\n");
				} 

				else {
					fprintf(STDOUT, "Too large cluster: %d, skip...", mark_global[k]);
					SplitLargeCluster(bitflags, row, col, type, dist, len, xsize, ysize, mark_global[k]);
					fprintf(STDOUT, "Done.\n");
				}
				if (bitflags[k] & GROUND) {											// Link the brach-cut to the image boundary
					DistToBorder(bitflags, row, col, type, len, xsize, ysize, mark_global[k]);
				}
			}
		}
	}
	free(G);
#if OUTPUTMST
	for (i = 0; i < xsize; ++i) {
		for (j = 0; j < ysize; ++j) {
			k = i*ysize+j;
			if(bitflags[k]&CUT && mark_global[k]==0) {
				mark_global[k] = -40000;
			}
		}
	}
	OpenFile(&fp, "mark_global", "wb");
	WriteData(fp, mark_global, xsize*ysize, sizeof(int));
	exit(-1);
#endif
	return bad_flag;
}

// for the boundary-balanced residue cluster, link the branch cuts to the nearest boundary
void DistToBorder(unsigned char *bitflags, short *row, short *col, int *type, 
				  const int len, const int xsize, const int ysize, int key_type) {

	int i, j, k, ka, min=xsize;
	char pos;
	int disttob[4];

	//
	for (ka = 0; ka < len; ++ka) {
		if (type[ka] == key_type) {
			disttob[0] = row[ka];
			disttob[1] = ysize-1-col[ka];
			disttob[2] = xsize-1-row[ka];
			disttob[3] = col[ka];
			for (k = 0; k < 4; ++k) {
				if (disttob[k]<min) {
					min = disttob[k];
					pos = "urdl"[k];
					i = row[ka];
					j = col[ka];
				}
			}
		}
	}
	switch (pos) {
	 						
	case 'u':
		PlaceCut(bitflags, i, j, 0, j, xsize, ysize);
		bitflags[j] |= CUT;
		break;
	case 'd':
		PlaceCut(bitflags, i, j, xsize-1, j, xsize, ysize);
		bitflags[(xsize-1)*ysize+j] |= CUT;
		break;
	case 'l':
		PlaceCut(bitflags, i, j, i, 0, xsize, ysize);
		bitflags[i*ysize] |= CUT;
		break;
	case 'r':
		PlaceCut(bitflags, i, j, i, ysize-1, xsize, ysize);
		bitflags[i*ysize+ysize-1] |= CUT;
		break;
	}
}

// global phase unwrapping through integration

void PathIntegAroundCuts(float *phase, short **unwrapped, unsigned char *bitflags, const int xsize, const int ysize, int bad_flag) {

	int i, j, k, ii, jj, kk, boxi, boxj;
	float delta;
	short delta_k;
	IntQueue *Q;
	int index;
	FILE *ofp;
	
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {

			if (bitflags[i*ysize+j] & CUT) {
				bitflags[i*ysize+j] &= ~NEED_TODO;
			} else {
				bitflags[i*ysize+j] |= NEED_TODO;
			}
		}
	}
	
	fprintf(STDOUT, "Path integration around cuts...\n");
	AllocateMemory(&Q, 1, sizeof(struct IntQueue_tag));
	InitIntQueue(Q);
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if (bitflags[k] & NEED_TODO) {
				EnIntQueue(Q, k);
				bitflags[k] &= ~NEED_TODO;
				unwrapped[i][j] = 0;
				
				while (Q->head != Q->tail) {
					index = DeIntQueue(Q);
					boxi = index/ysize;
					boxj = index%ysize;
					for (ii = boxi-1; ii <= boxi+1; ii++) {
						for (jj = boxj-1; jj <= boxj+1; jj++) {
							if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {	//
								continue;
							}
							if (ii==boxi || jj==boxj) {						//
								kk = ii*ysize+jj;
								if (bitflags[kk]&NEED_TODO) {
									delta = phase[kk]-phase[index];
									delta_k = 0;
									if (delta >= PI) {
										delta_k = -1;
									} else if (delta <= -PI) {
										delta_k = 1;
									}
									EnIntQueue(Q, kk);
									bitflags[kk] &= ~NEED_TODO;
									unwrapped[ii][jj] = unwrapped[boxi][boxj]+delta_k;
								}
							}
						} //end of jj...
					} //end of ii...
				} //end of while...
				if (bad_flag == FALSE) {
					goto solve_cut;
				}
				//goto solve_cut;
			} //end of if...
		} //end of j...
	} //end of i...
	

solve_cut:
	for (i = 0; i < xsize; ++i) {
		for (j = 0; j < ysize; ++j) {
			k = i*ysize+j;
			if (bitflags[k]&CUT) {

				if (j-1>0 && unwrapped[i][j-1]!=0) {
					unwrapped[i][j] = unwrapped[i][j-1];
				} else if (i-1>0 && unwrapped[i-1][j]!=0) {
					unwrapped[i][j] = unwrapped[i-1][j];
				} else {
					phase[k] = 0.0f;
				}

				phase[k] = 0.0f;
				unwrapped[i][j] = 0;
			} else {
				phase[k] += unwrapped[i][j]*2*PI;
			}
		}
	}
	DestroyIntQueue(Q);
	free(Q);
	fprintf(STDOUT, "Done.\n");
}

// for those residue clusters whose number of residues is greater than the threshold, a mask is generated
void SplitLargeCluster(unsigned char *bitflats, short *row, short *col, int *type, float *dist,
					   const int len, const int xsize, const int ysize, int key_type) {

	int i, ii, jj;
	int cur_dist;
	int max_row=0, max_col=0, min_row=xsize, min_col=ysize;

	for (i = 0; i < len; ++i) {
		if (type[i] == key_type) {
			if (row[i] > max_row)	max_row = row[i];
			if (row[i] < min_row)	min_row = row[i];
			if (col[i] > max_col)	max_col = col[i];
			if (col[i] < min_col)	min_col = col[i];
		}
	}
	for (i = 0; i < len; ++i) {
		if (type[i] == key_type) {
			cur_dist = (int)(dist[i]+1);
			for (ii = row[i]-cur_dist; ii <= row[i]+cur_dist; ++ii) {
				for (jj = col[i]-cur_dist; jj <= col[i]+cur_dist; ++jj) {
					if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {
						continue;
					}
					if (DIS(ii, jj, row[i], col[i])<=dist[i]) {
						if (ii<=max_row && ii>=min_row && jj<=max_col && jj>=min_col) {
							bitflats[ii*ysize+jj] |= CUT;
							bitflats[ii*ysize+jj] &= ~NEED_TODO;
						}
					}
				} //end of jj...
			} //end of ii...
		} //end of if...
	} //end of i...
}


// for those residue clusters whose number of residues is greater than the threshold, block them, 
//  ... solve MST for each block, and then stitch
void GenMSTforLargeCluster(unsigned char *bitflags, short *row, short *col, int *type, float *dist, 
						   const int len, const int xsize, const int ysize, const int cluster_limit, const int key_type) {

	int		i, k, flag=0, index;
	short	max_row=0, min_row=xsize, max_col=0, min_col=ysize, cur_row, cur_col;
	int		count=0, cut_num, piece_num, step;
	short	*key_row, *key_col;
	float	*key_dist;
	int		*key_kind, kind=0;
	MGraph	*G;
	int		status, u;

	for (k = 0; k < len; ++k) {
		if (type[k] == key_type) {
			++count;
			if (row[k] > max_row)	max_row = row[k];
			if (row[k] < min_row)	min_row = row[k];
			if (col[k] > max_col)	max_col = col[k];
			if (col[k] < min_col)	min_col = col[k];
		}
	}
	
	AllocateMemory(&key_row, count, sizeof(short));
	AllocateMemory(&key_col, count, sizeof(short));
	AllocateMemory(&key_dist, count, sizeof(float));
	AllocateMemory(&key_kind, count, sizeof(int));
	memset(key_kind, 0, count*sizeof(int));

	//
	for (i = 0, k = 0; k < len; ++k) {
		if (type[k] == key_type) {
			key_row[i] = row[k];
			key_col[i] = col[k];
			key_dist[i++] = dist[k];
		}
	}

	//
	cut_num = (float)count*2/MAX_CLUSTER;
	flag = (max_row-min_row)>(max_col-min_col) ? 0 : 1;

	fprintf(STDOUT, "%d residues, total:%d pieces\n", count, cut_num);
	G = (MGraph *)malloc(sizeof(MGraph));


	//
	if (flag == 0) {

		//
		step = (max_row-min_row)/cut_num+1;
		//
		kind = 1;
		//
		for (i = 0; i < count; ++i) {
			if (key_row[i] == min_row) {
				u = key_row[i]*ysize+key_col[i];
				break;
			}
		}
		//
		for (k = min_row; k <= max_row; k+=step) {

			//
			piece_num = 0;
			for (i = 0; i < count; ++i) {

				if (key_row[i]>=k && key_row[i]<k+step) {
					key_kind[i] = kind;
					piece_num++;
				}
			}
			fprintf(STDOUT, "%d residues for kid-cluster %d, start at (%d,%d)\n", piece_num, kind, u/ysize, u%ysize);

			// MST
			status = CreateGraph(G, key_row, key_col, key_kind, key_dist, count, xsize, ysize, cluster_limit, kind);
			if (status) {
				MinSpanTreePrim(G, u, bitflags, xsize, ysize);	// generate MST and place branch cuts
				DestroyGraph(G);
			} else {
				fprintf(STDOUT, "Large block after cut, skip...\n");

			}
			
			//
			cur_row = -1;
			for (i = 0; i < count; ++i) {
				if (key_kind[i] == kind) {
					if (key_row[i]>cur_row) {
						cur_row = key_row[i];
						u = key_row[i]*ysize+key_col[i];
						index = i;
					}
				}
			}
			++kind;
			key_kind[index] = kind;
		}
	} else if (flag == 1) {

		step = (max_col-min_col)/cut_num+1;
		kind = 1;
		for (i = 0; i < count; ++i) {
			if (key_col[i] == min_col) {
				u = key_row[i]*ysize+key_col[i];
				break;
			}
		}
		for (k = min_col; k <= max_col; k+=step) {

			piece_num = 0;
			for (i = 0; i < count; ++i) {

				if (key_col[i]>=k && key_col[i]<k+step) {
					key_kind[i] = kind;
					piece_num++;
				}
			}
			fprintf(STDOUT, "%d residues for kid-cluster %d, start at (%d,%d)\n", piece_num, kind, u/ysize, u%ysize);
			status = CreateGraph(G, key_row, key_col, key_kind, key_dist, count, xsize, ysize, cluster_limit, kind);
			if (status) {
				MinSpanTreePrim(G, u, bitflags, xsize, ysize);					// generate MST and place branch cuts
				DestroyGraph(G);
			} else {
				fprintf(STDOUT, "Large block after cut, skip...\n");
			}
			
			cur_col = -1;
			for (i = 0; i < count; ++i) {
				if (key_kind[i] == kind) {
					if (key_col[i]>cur_col) {
						cur_col = key_col[i];
						u = key_row[i]*ysize+key_col[i];
						index = i;
					}
				}
			}
			++kind;
			key_kind[index] = kind;
		}
	}

	free(key_row);
	free(key_col);
	free(key_dist);
	free(key_kind);
	free(G);
}


void FindLargeKind(unsigned char *bitflags, int *mark_global, const int xsize, const int ysize, const int resnum) {

	int i, j, k, ii, jj, kk, boxi, boxj, ka, cur_type=1;
	int *active_list, active_num=0;
	int max_kind_num=0, max_kind=0;
	
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			bitflags[i*ysize+j] &= ~VISITED;
			bitflags[i*ysize+j] &= ~ACTIVE;
			if (mark_global[i*ysize+j]==0) {
				bitflags[i*ysize+j] |= VISITED;
			}
		}
	}
	
	fprintf(STDOUT, "\nGet biggest kind block...\n");
	AllocateMemory(&active_list, xsize*ysize, sizeof(int));
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			
			k = i*ysize+j;
			if ((bitflags[k]&VISITED) && mark_global[k]==0) {
				
				active_num = 0;
				active_list[active_num++] = k;
				bitflags[k] |= ACTIVE;
				mark_global[k] = -cur_type;
				for (ka = 0; ka < active_num; ka++) {
					boxi = active_list[ka]/ysize;
					boxj = active_list[ka]%ysize;
					for (ii = boxi-1; ii <= boxi+1; ii++) {
						for (jj = boxj-1; jj<=boxj+1; jj++) {
							
							if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {
								continue;
							}
							kk = ii*ysize+jj;
							if ((bitflags[kk]&VISITED) && !(bitflags[kk]&ACTIVE)) {
								active_list[active_num++] = kk;
								bitflags[kk] |= ACTIVE;
								mark_global[kk] = -cur_type;
							}
						}
					} //end of ii...
				} //end of ka...
				
				for (ka = 0; ka < active_num; ka++) {
					bitflags[active_list[ka]] &= ~ACTIVE;
					bitflags[active_list[ka]] &= ~VISITED;
				}
				if (active_num > max_kind_num) {
					max_kind_num = active_num;
					max_kind = -cur_type;
				}
				cur_type++;
			} //end of if...
		}
	} //end of i...
	free(active_list);
	
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if (mark_global[k]<0 && mark_global[k]!=max_kind) {
				bitflags[k] &= ~VISITED;
			} else if (mark_global[k] == max_kind) {
				bitflags[k] |= VISITED;
			}
		}
	}
	fprintf(STDOUT, "Done.\n");
}

void SolveLargeKind(float *phase, short **unwrapped, unsigned char *bitflags, const int xsize, const int ysize) {

	int i, j, k, ii, jj, kk, boxi, boxj;
	float delta;
	short delta_k;
	IntQueue *Q;
	int index;
	
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			
			if (bitflags[i*ysize+j]&CUT || !(bitflags[i*ysize+j]&VISITED)) {
				bitflags[i*ysize+j] &= ~NEED_TODO;
			} else {
				bitflags[i*ysize+j] |= NEED_TODO;
			}
		}
	}
	
	fprintf(STDOUT, "Path integration around cuts...\n");
	AllocateMemory(&Q, 1, sizeof(struct IntQueue_tag));
	InitIntQueue(Q);
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if (bitflags[k] & NEED_TODO) {
				EnIntQueue(Q, k);
				bitflags[k] &= ~NEED_TODO;
				unwrapped[i][j] = 0;
				
				while (Q->head != Q->tail) {
					index = DeIntQueue(Q);
					boxi = index/ysize;
					boxj = index%ysize;
					for (ii = boxi-1; ii <= boxi+1; ii++) {
						for (jj = boxj-1; jj <= boxj+1; jj++) {
							if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {	//
								continue;
							}
							if (ii==boxi || jj==boxj) {						//
								kk = ii*ysize+jj;
								if (bitflags[kk]&NEED_TODO) {
									delta = phase[kk]-phase[index];
									delta_k = 0;
									if (delta >= PI) {
										delta_k = -1;
									} else if (delta <= -PI) {
										delta_k = 1;
									}
									EnIntQueue(Q, kk);
									bitflags[kk] &= ~NEED_TODO;
									unwrapped[ii][jj] = unwrapped[boxi][boxj]+delta_k;
								}
							}
						} //end of jj...
					} //end of ii...
				} //end of while...
				goto solve_cut;
			} //end of if...
		} //end of j...
	} //end of i...
	
	
solve_cut:
	for (i = 0; i < xsize; ++i) {
		for (j = 0; j < ysize; ++j) {
			k = i*ysize+j;
			if (bitflags[k]&CUT) {
				
				if (j-1>0 && unwrapped[i][j-1]!=0) {
					unwrapped[i][j] = unwrapped[i][j-1];
				} else if (i-1>0 && unwrapped[i-1][j]!=0) {
					unwrapped[i][j] = unwrapped[i-1][j];
				} else {
					phase[k] = 0.0f;
				}
			}
			if (!(bitflags[k]&VISITED)) {
				phase[k] = 0.0f;
			}
			phase[k] += unwrapped[i][j]*2*PI;
		}
	}
	DestroyIntQueue(Q);
	free(Q);
	fprintf(STDOUT, "Done.\n");
}
