/*************************************************************************************************************************/
/* Description:
/* This program is used to complete the residue clustering.                                                              
/* This cluster algorithm can be used for the ODLS method[1], FLS method[2] and PU method in [3].                        
/* [1] H. Yu, Z. Li, and Z. Bao, "Residues cluster-based segmentation and outlier-detection method for large-scale phase 
/*     unwrapping," IEEE Trans. Image. Processing., vol. 20, no. 10, pp. 2865-2875, Oct. 2011.                         
/* [2] H. Yu, M. Xing, and Z. Bao, "A fast phase unwrapping method for large-scale interferograms," IEEE Trans. Geosci. 
/*     Remote Sens., vol. 51, no. 7, pp. 4240-4248, Jul. 2013.                                                           
/* [3] H. Yu, Y. Lan, J. Xu, and D. An, "Large-scale L0-norm and L1-norm 2-D phase unwrapping," IEEE Trans. Geosci. 
/*     Remote Sens., vol. 55, no. 8, pp. 4712-4728, Aug. 2017.                                                           

/* Release date of FLSv1.0: Aug.05, 2018
/* Latest updated date of FLSv2.0: Oct.02, 2018
/*************************************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cluster.h"
#include "file.h"
#include "util.h"

#define DEBUG_FLG	0	
#define DIST_OUT	0	

/* Variable Description ***************************************************************************************************
 * mark_global:    matrix for recording the cluster number of each residue in global
 * bitflags:       matrix for recording information of residues, branch-cut and ground
 * qual:           matrix of correlation coefficient
 * xsize,ysize:    the size of input data 
 * resnum:         the total number of residues.				
 * delta:          the step size of cluster radii of residues
 * lowdis,highdis: lower and upper limits of the initial value of cluster radii
 * row,col:        vectors for recording the position(row and column) of all residues
 * type:           vector for recording the residue cluster number
 * dist_in:        vector for recording the final cluster radii of each residue
 */
/*************************************************************************************************************************/

void ClusterRepeatErgodic(int **mark_global, unsigned char *bitflags, float *qual, const int xsize, const int ysize, 
						  int resnum, float delta, float lowdis, float highdis, short **row, short **col, int **type, float **dist_in) {

	int *active_list, *merg_list;												// current residue cluster list, residue cluster list to merge residues

	int i, j, k, ka, kb, ii, jj, kk, boxi, boxj, boxiii, boxjjj, iii, jjj, kkk;	// loop variables
	int cur_dist, active_num, merg_num=0;										// window radius, the length of active_list
	int charge, ground_flag;													// polarity of residues in one cluster, a flag indicating whether cluster is grounded
	int max_type = 1, temp_type;												// dynamic number of residue cluster
	int c_times = 1, surplus;													// the times of clustering, number of residues remaining after clustering
	
	float corr, corr_min = 1.0, corr_max = 0.0, dist_k;							// the parameter used to calculate the initial value of the cluster radii
	short *dist;																// cluster radii matrix
	float temp_dist=0.0f;														// temporary variable used to increase the cluster radii.

#if DIST_OUT
	FILE *dfp;
#endif

	AllocateMemory(&dist, xsize*ysize, sizeof(short));
	memset(dist, 0, xsize*ysize*sizeof(short));

	// calculate the maximum and minimum values of the coherence coefficient.
	for (i = 0; i < xsize-1; i++) {
		for (j = 0; j < ysize-1; j++) {
			
			k = i*ysize+j;
			corr = (qual[k] + qual[k+1] + qual[k+ysize] + qual[k+ysize+1]) / 4;
			if (corr > corr_max) {
				corr_max = corr;
			}
			if (corr < corr_min) {
				corr_min = corr;
			}
		}
	}
	
	//  calculate matrix dist (i.e., cluster radii of each residue)
	for (i = 0; i < xsize-1; i++) {
		for (j = 0; j < ysize-1; j++) {
			
			k = i*ysize+j;
			if (bitflags[k] & RESIDUE) {
				bitflags[k] |= NEED_TODO;
				corr = (qual[k] + qual[k+1] + qual[k+ysize] + qual[k+ysize+1]) / 4;
				dist[k] = (short)(((highdis-lowdis)/(corr_min-corr_max))*corr + (((lowdis*corr_min)-(highdis*corr_max))/(corr_min-corr_max))+0.5);
			}
		}
	}

	//
	free(qual);
	AllocateMemory(mark_global, xsize*ysize, sizeof(int));
 	memset((*mark_global), 0, xsize*ysize*sizeof(int));
	AllocateMemory(&active_list, resnum+1, sizeof(int));
	memset(active_list, 0, (resnum+1)*sizeof(int));
	AllocateMemory(&merg_list, resnum, sizeof(int));
	memset(merg_list, 0, resnum*sizeof(int));

	// clustering
	while (TRUE) {
		fprintf(STDOUT, "%d time cluster...\n", c_times);
		for (i = 0; i < xsize-1; i++) {								// Global traversal to find residues
			fprintf(STDOUT, "\b\b\b\b%d%%", i*100/(xsize-2));
			for (j = 0; j < ysize-1; j++){
				
				k = i*ysize+j;
				if ((bitflags[k]&NEED_TODO) && (bitflags[k]&RESIDUE) && (*mark_global)[k]==0) {		// clustering if the residue is found
					
					charge = bitflags[k]&POS_RES ? +1 : -1;
					active_num = 0;
					active_list[active_num++] = k;
					bitflags[k] |= ACTIVE;
					(*mark_global)[k] = max_type;
					ground_flag = FALSE;

					for (ka = 0; ka < active_num; ka++) {			// traverse the list of residues
						if (bitflags[active_list[ka]] & VISITED) {
							continue;
						}
						boxi = active_list[ka] / ysize;
						boxj = active_list[ka] % ysize;
						bitflags[active_list[ka]] |= VISITED;
						cur_dist = (int)(highdis + (c_times-1)*delta + (float)dist[active_list[ka]]);
						for (ii = boxi-cur_dist; ii <= boxi+cur_dist; ii++) {	    //search residues in the window that can be clustered
							for (jj = boxj-cur_dist; jj <=boxj+cur_dist; jj++) {

								if (ii<0 || ii >= xsize || jj<0 || jj >= ysize) {
									continue;
								}
								kk = ii*ysize+jj;
								dist_k = (float)DIS(boxi, boxj, ii, jj);
								if (dist_k <= dist[active_list[ka]]+dist[kk]) {
									if (ii==0 || ii==xsize-1 || jj==0 || jj==ysize-1) {
										if (ground_flag == FALSE) {
											ground_flag = TRUE;
											for (iii = 0; iii < active_num; iii++) {
												bitflags[active_list[iii]] |= GROUND;
											}
										}
									}
									if ((*mark_global)[kk]!=0 && (*mark_global)[kk]!=max_type) {
										// if the type-number of the residue kk is x, all residues with type-number x are set to the current type-number
										
										if (bitflags[kk] & GROUND) {		// if the residue cluster is ground balanced
											ground_flag = TRUE;
											for (iii = 0; iii < active_num; iii++) {
												bitflags[active_list[iii]] |= GROUND;
											}
										}
										temp_type = (*mark_global)[kk];
										merg_num = 0;
										merg_list[merg_num++] = kk;
										for (kb = 0; kb < merg_num; kb++) {
											
											boxiii = merg_list[kb] / ysize;
											boxjjj = merg_list[kb] % ysize;
											for (iii = boxiii-cur_dist; iii <= boxiii+cur_dist; iii++) {
												for (jjj = boxjjj-cur_dist; jjj <= boxjjj+cur_dist; jjj++) {
													
													if (iii<0 || iii>=xsize || jjj<0 || jjj>=ysize) {
														continue;
													}
													kkk = iii*ysize+jjj;
													if ((*mark_global)[kkk] == temp_type) {
														(*mark_global)[kkk] = max_type;
														active_list[active_num++] = kkk;
														bitflags[kkk] |= ACTIVE;
														merg_list[merg_num++] = kkk;
														if (ground_flag) {
															bitflags[kkk] |= GROUND;
														}
													}
												}
											}
										}
									} else if (bitflags[kk]&RESIDUE && !(bitflags[kk]&ACTIVE)) {
										//  if the residue has no type-number, merge the residue into current cluster

										charge += bitflags[kk]&POS_RES ? (+1) : (-1);
										active_list[active_num++] = kk;
										bitflags[kk] |= ACTIVE;
										(*mark_global)[kk] = max_type;
										if (ground_flag) {
											bitflags[kk] |= GROUND;
										}
									}
								}
							} //end of jj...
						} //end of ii...
					} //end of search list...
					if (charge!=0 && ground_flag==FALSE) {						//if the residue cluster is unbalanced, increase the cluster radii
						for (ka = 0; ka < active_num; ka++) {
							(*mark_global)[active_list[ka]] = 0;
							bitflags[active_list[ka]] &= ~NEED_TODO;
							
						}
					} else {													//if the residue cluster is balanced, increase the number of residue cluster
						++max_type;
					}
					for (ka = 0; ka < active_num; ka++) {						//
						bitflags[active_list[ka]] &= ~ACTIVE;
					}
				} //end of if RESIDUE
			} //end of j...
		} //end of i...

				// looking for unbalanced residues, increase their cluster radii and determine if clustering should be end
		for (surplus = 0, iii = 0; iii < xsize-1; iii++) {
			for (jjj = 0; jjj < ysize-1; jjj++) {
				kkk = iii*ysize+jjj;
				if (bitflags[kkk]&RESIDUE && (*mark_global)[kkk]==0) {
					temp_dist = (float)dist[kkk]+delta;
					dist[kkk] = (short)(temp_dist+0.5);
					//dist[kkk] = dist[kkk]*2;
					++surplus;
					bitflags[kkk] &= ~VISITED;
					bitflags[kkk] |= NEED_TODO;
				}
			}
		}
		if (surplus!=0) {
			fprintf(STDOUT, "\n%d residues unbanlance.\n", surplus);
			c_times++;
		} else {
			break;
		}
	}
	free(active_list);
	free(merg_list);

#if DIST_OUT
	OpenFile(&dfp, "dist", "wb");
	WriteData(dfp, dist, xsize*ysize, sizeof(short));
#endif

	AllocateMemory(row, resnum, sizeof(short));
	AllocateMemory(col, resnum, sizeof(short));
	AllocateMemory(type, resnum, sizeof(int));
	AllocateMemory(dist_in, resnum, sizeof(float));
	active_num = 0;
	for (i = 0; i < xsize-1; i++) {
		for (j = 0; j < ysize-1; j++) {
			k = i*ysize+j;
			if ((*mark_global)[k]!=0) {
				(*row)[active_num] = i;
				(*col)[active_num] = j;
				(*type)[active_num] = (*mark_global)[k];
				(*dist_in)[active_num++] = dist[k];
			}
		}
	}
	free(dist);

	fprintf(STDOUT, "\nSort and rekind cluster...\n");
	LinearSortNew(*row, *col, *type, *dist_in, resnum);
	ReKind(*type, resnum);

	for (ka = 0; ka < resnum; ka++) {
		(*mark_global)[(*row)[ka]*ysize+(*col)[ka]] = (*type)[ka];
	}
}

// single clustering, need to be checked
void Cluster(int *mark_global, unsigned char *bitflags, float *qual, const int xsize, const int ysize, int resnum, 
			 float delta, float lowdis, float highdis, short **row, short **col, int **type, float **dist_in) {

	float corr, corr_min = 1.0, corr_max = 0.0, dist_k;
	float *dist;
	short *add_cnt, cur_add_dist_cnt;			//
	
	int cur_dist;
	int i, j, k, ka, ii, jj, kk, boxi, boxj, iii, jjj, kkk;
	int charge;
	int active_num;
	int max_type=1, ground_flag, cur_type;

	int *active_list;
#if DEBUG_FLG
	int debug_flag;
	FILE *tfp;
#endif
#if DIST_OUT
	FILE *dfp;
#endif

	AllocateMemory(&active_list, resnum+1, sizeof(int));
	AllocateMemory(&dist, xsize*ysize, sizeof(float));
	AllocateMemory(&add_cnt, xsize*ysize, sizeof(short));
	memset(dist, 0, xsize*ysize*sizeof(float));
	memset(active_list, 0, (resnum+1)*sizeof(int));
	memset(add_cnt, 0, xsize*ysize*sizeof(short));

	// calculate the maximum and minimum values of the coherence coefficient.
	for (i = 0; i < xsize-1; i++) {
		for (j = 0; j < ysize-1; j++) {

			k = i*ysize+j;
			corr = (qual[k] + qual[k+1] + qual[k+ysize] + qual[k+ysize+1]) / 4;
			if (corr > corr_max) {
				corr_max = corr;
			}
			if (corr < corr_min) {
				corr_min = corr;
			}
		}
	}

	// calculate matrix dist (i.e., cluster radii of each residue)
	for (i = 0; i < xsize-1; i++) {
		for (j = 0; j < ysize-1; j++) {

			k = i*ysize+j;
			if (bitflags[k] & RESIDUE) {
				corr = (qual[k] + qual[k+1] + qual[k+ysize] + qual[k+ysize+1]) / 4;
				dist[k] = ((highdis-lowdis)/(corr_min-corr_max))*corr + (((lowdis*corr_min)-(highdis*corr_max))/(corr_min-corr_max));
			}
		}
	}

	free(qual);
	//clustering
	for (i = 0; i < xsize-1; i++) {
		fprintf(STDOUT, "\b\b\b\b%d%%", i*100/(xsize-2));
		for (j = 0; j < ysize-1; j++) {

			k = i*ysize+j;
			if ((bitflags[k]&RESIDUE) && mark_global[k]==0) {		
				// For the residue i, only after getting the balanced cluster, the new clustering of a unclustered residue j can be started.

#if DEBUG_FLG
				if (max_type == 6093) {
					debug_flag = TRUE;
					OpenFile(&tfp, "debugcluster", "wb");
					WriteData(tfp, mark_golbal, xsize*ysize, sizeof(int));
					OpenFile(&tfp, "debugdist", "wb");
					WriteData(tfp, dist, xsize*ysize, sizeof(float));
					exit(DEBUG_FLG);
				}
#endif
				//ACTIVE indicates whether the residue is in the active_list
				charge = bitflags[k] & POS_RES ? 1 : -1;
				active_num = 0;
				active_list[active_num++] = k;
				bitflags[k] |= ACTIVE;
				ground_flag = FALSE;
				cur_dist = (int)(highdis+1);

				while (TRUE) {										// if the current cluster is unbalanced, continue to add new residue

					for (ka = 0; ka < active_num; ka++) {			//
						if (bitflags[active_list[ka]] & VISITED) {	//
							continue;
						}
						boxi = active_list[ka]/ysize;
						boxj = active_list[ka]%ysize;
						mark_global[active_list[ka]] = max_type;
						bitflags[active_list[ka]] |= VISITED;		//

						//
						//
						for (ii = boxi-cur_dist; ii <= boxi+cur_dist; ii++) {
							for (jj = boxj-cur_dist; jj <= boxj+cur_dist; jj++) {					//for each residue in list, add a window to clustering
								kk = ii*ysize+jj;
								if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {						//
									continue;
								} 
								dist_k = (float)DIS(ii, jj, boxi, boxj);
								if (dist_k < dist[active_list[ka]] || dist_k < dist[kk]) {			//

									if (ii==0 || ii==xsize-1 || jj==0 || jj==ysize-1) {				//if reach the boundary of image, ground the cluster
										ground_flag = TRUE;
										bitflags[active_list[ka]] |= GROUND;
									} 
									if (mark_global[kk]!=0 && mark_global[kk]!=max_type) {			
										//if this residue belongs to another cluster, merge it into the current cluster

										if (bitflags[kk] & GROUND) {
											ground_flag = TRUE;
											bitflags[active_list[ka]] |= GROUND;
											for (iii = 0; iii < active_num; iii++) {
												bitflags[active_list[iii]] |= GROUND;
											}
										}

										cur_type = mark_global[kk];
										for (iii = 0; iii < xsize; iii++) {
											for (jjj = 0; jjj < ysize; jjj++) {
												kkk = iii*ysize+jjj;
												if (mark_global[kkk] == cur_type) {					//update the type-number
													mark_global[kkk] = max_type;
													active_list[active_num++] = kkk;				//
													bitflags[kkk] |= ACTIVE;
												}
											}
										}
									} else if ((bitflags[kk]&RESIDUE) && !(bitflags[kk]&ACTIVE)) {	// if the residue is a new one
										
										charge += bitflags[kk] & POS_RES ? 1 : -1;
										active_list[active_num++] = kk;
										bitflags[kk] |= ACTIVE;
										mark_global[kk] = max_type;
										if (ground_flag) {
											bitflags[kk] |= GROUND;
										}
									}
								}
							} //end of jj...
						} //end of ii...
					} //end of search active_list...
					if (charge!=0 && ground_flag==FALSE) {							// if unbalanced after traversing the list
						cur_add_dist_cnt = add_cnt[active_list[0]];					// record the residue with the least increase in cluster radii
						for (ka = 1; ka < active_num; ka++) {
							if (add_cnt[active_list[ka]] < cur_add_dist_cnt) {
								cur_add_dist_cnt = add_cnt[active_list[ka]];
							}
						}
						for (ka = 0; ka < active_num; ka++) {						//
							if (add_cnt[active_list[ka]] == cur_add_dist_cnt) {		// if it is the residue with the least increase in cluster radii
								dist[active_list[ka]] += delta;						// increase its cluster radii
								add_cnt[active_list[ka]]++;
								bitflags[active_list[ka]] &= ~VISITED;				// change the traversal flag for adding window so that it can be manipulated in the next time
							}
						}
						cur_dist += delta;		//

					} else {
						break;
					}
				} //end of while...

				++max_type;											// the clustering for the residue i is finished, and the type-number is be increased
				for (ka = 0; ka < active_num; ka++) {			    	// Remove these residues(i.e., the cluster contains residue i) from the list
					bitflags[active_list[ka]] &= ~ACTIVE;
				}
			} //end of if RESIDUE...
		} //end of j
	} //end of i

	free(active_list);
	free(add_cnt);

#if DIST_OUT
	OpenFile(&dfp, "dist", "wb");
	WriteData(dfp, dist, xsize*ysize, sizeof(float));
#endif
	AllocateMemory(row, resnum, sizeof(short));
	AllocateMemory(col, resnum, sizeof(short));
	AllocateMemory(type, resnum, sizeof(int));
	AllocateMemory(dist_in, resnum, sizeof(float));
	active_num = 0;
	for (i = 0; i < xsize-1; i++) {
		for (j = 0; j < ysize-1; j++) {
			k = i*ysize+j;
			if (mark_global[k]!=0) {
				(*row)[active_num] = i;
				(*col)[active_num] = j;
				(*type)[active_num] = mark_global[k];
				(*dist_in)[active_num++] = dist[k];
			}
		}
	}
	free(dist);

	fprintf(STDOUT, "\nSort and rekind cluster...\n");
	LinearSortNew(*row, *col, *type, *dist_in, resnum);
	ReKind(*type, resnum);
	
	for (ka = 0; ka < resnum; ka++) {
		mark_global[(*row)[ka]*ysize+(*col)[ka]] = (*type)[ka];
	}
}



// find the area of each residue cluster and set its type-number to the residue cluster
void SplitCluster(int *mark_global, short *row, short *col, int *type, float *dist, const int len, const int xsize, const int ysize) {

	int i, ka, ii, jj, kk, boxi, boxj;
	int cur_dist;
	int *four_neighbor;

	fprintf(STDOUT, "Split cluster...\n");
	AllocateMemory(&four_neighbor, 4*2, sizeof(int));
	for (i = 0; i < len; i++) {
		fprintf(STDOUT, "\b\b\b\b%d%%", i*100/(len-1));
		four_neighbor[0] = row[i];
		four_neighbor[1] = col[i];
		four_neighbor[2] = row[i]+1;
		four_neighbor[3] = col[i];
		four_neighbor[4] = row[i];
		four_neighbor[5] = col[i]+1;
		four_neighbor[6] = row[i]+1;
		four_neighbor[7] = col[i]+1;
		cur_dist = (int)(dist[i]+1);
		for (ka = 0; ka < 3; ka++) {
			
			boxi = four_neighbor[ka*2];
			boxj = four_neighbor[ka*2+1];
			for (ii = boxi-cur_dist; ii <= boxi+cur_dist; ii++) {
				for (jj = boxj-cur_dist; jj <= boxj+cur_dist; jj++) {
					kk = ii*ysize+jj;
					if (ii<0 || ii>xsize-1 || jj<0 || jj>ysize-1) {
						continue;
					}
					if (DIS(ii, jj, boxi, boxj) < dist[i]) {
						mark_global[kk] = type[i];
					}
				}
			}
		}
	}
	free(four_neighbor);
	fprintf(STDOUT, "\nDone.\n");
}

// find the biggest good area, set it's type-number to -1
int GetKindBlock(int *mark_global, unsigned char *bitflags, const int xsize, const int ysize, const int resnum) {

	int i, j, k, ii, jj, kk, boxi, boxj, ka, cur_type=1;
	int *active_list, active_num=0;
	int max_kind_num=0, max_kind=0;

	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			bitflags[i*ysize+j] &= ~VISITED;
			bitflags[i*ysize+j] &= ~ACTIVE;
			if (mark_global[i*ysize+j]==0) {
				bitflags[i*ysize+j] |= NEED_TODO;
			}
		}
	}

	fprintf(STDOUT, "Get kind block...\n");
	AllocateMemory(&active_list, xsize*ysize-resnum+10, sizeof(int));
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {

			k = i*ysize+j;
			if ((bitflags[k]&NEED_TODO) && mark_global[k]==0) {

				active_num = 0;
				active_list[active_num++] = k;
				bitflags[k] |= ACTIVE;
				mark_global[k] = -cur_type;
				for (ka = 0; ka < active_num; ka++) {
					boxi = active_list[ka]/ysize;
					boxj = active_list[ka]%ysize;
					for (ii = boxi-1; ii <= boxi+1; ii++) {
						for (jj = boxj-1; jj<=boxj+1; jj++) {
							
							if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {
								continue;
							}
							kk = ii*ysize+jj;
							if ((bitflags[kk]&NEED_TODO) && !(bitflags[kk]&ACTIVE)) {
								active_list[active_num++] = kk;
								bitflags[kk] |= ACTIVE;
								mark_global[kk] = -cur_type;
							}
						}
					} //end of ii...
				} //end of ka...

				for (ka = 0; ka < active_num; ka++) {
					bitflags[active_list[ka]] &= ~ACTIVE;
				}
				if (active_num > max_kind_num) {
					max_kind_num = active_num;
					max_kind = -cur_type;
				}
				cur_type++;
			} //end of if...
		}
	} //end of i...
	free(active_list);

	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if (mark_global[k]<0 && mark_global[k]!=max_kind) {
				mark_global[k] = 0;
			} else if (mark_global[k] == max_kind) {
				mark_global[k] = -1;
			}
		}
	}
	fprintf(STDOUT, "Done.\n");
	return -1;
}

// find good area except the biggist one and give them type-number
void GetKindSmallerBlock(int *mark_global, unsigned char *bitflags, const int xsize, const int ysize, 
						 const int resnum, const int threshold) {

	int i, j, k, ka, ii, jj, kk, boxi, boxj;
	int *active_list, active_num=0;
	int cur_type=2;
	int max_kind, max_kind_num;

	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			if (mark_global[i*ysize+j]==0) {
				bitflags[i*ysize+j] |= NEED_TODO;
			} else {
				bitflags[i*ysize+j] &= ~NEED_TODO;
			}
		}
	}

	fprintf(STDOUT, "Get smaller kind block, upper limit is %d\n", threshold);
	AllocateMemory(&active_list, xsize*ysize-resnum, sizeof(int));
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if ((bitflags[k]&NEED_TODO) && mark_global[k]==0) {
				active_num = 0; 
				active_list[active_num++] = k;
				bitflags[k] |= ACTIVE;
				mark_global[k] = -cur_type;
				for (ka = 0; ka < active_num; ka++) {

					boxi = active_list[ka]/ysize;
					boxj = active_list[ka]%ysize;
					for (ii = boxi-1; ii <= boxi+1; ii++) {
						for (jj = boxj-1; jj<= boxj+1; jj++) {
							
							if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {
								continue;
							}
							kk = ii*ysize+jj;
							if ((bitflags[kk]&NEED_TODO) && !(bitflags[kk]&ACTIVE)) {
								active_list[active_num++] = kk;
								bitflags[kk] |= ACTIVE;
								mark_global[kk] = -cur_type;
							}
						}
					} //end of ii
				} //end of ka

				for (ka = 0; ka < active_num; ka++) {
					bitflags[active_list[ka]] &= ~ACTIVE;
				}
				if (active_num > max_kind_num) {
					max_kind_num = active_num;
					max_kind = -cur_type;
				}
				cur_type++;
			} //end of if
		}
	}
	free(active_list);
	fprintf(STDOUT, "Done.\n");
}
