/****************************************************************************************************************************/
/* Description:

/* FLS is a fast phase unwrapping(PU) method for large-scale interferogram with single baseline.
/* For the detailed information about the input, output and parameters of this code, please refer to the yu.cfg files.
/* For more information about the algorithm, please refer to [1]
/* [1] H. Yu, M. Xing, and Z. Bao, "A fast phase unwrapping method for large-scale interferograms," IEEE Trans. Geosci. 
       Remote Sens., vol. 51, no. 7, pp. 4240-4248, Jul. 2013. 

/* Release date of FLSv1.0: Aug.05, 2018
/* Latest updated date of FLSv2.0: Oct.02, 2018
/* Modify:
/*   1. Modified the function used to generate the mask of large cluster.
/*   2. Fixed the bug that would create artificial lines in PU result under certain paramethers
*****************************************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "util.h"
#include "file.h"
#include "qual.h"
#include "residue.h"
#include "cluster.h"
#include "brcut.h"
#include "solver.h"

#define ONECLUSTER	0   //
#define OTHERWORK	1   // 0: Only perform residue clustering;  1: perform residue clustering and then phase unwrapping
#define BIGKIND		0   // 0: Solve both the high quality area and the noise area

int main(int argc, char *argv[]) {

	int		xsize, ysize;					//read the size of the data from the configuration file

	float	delta, lowdis, highdis;			//read clustering parameters from the configuration file
	int		max_cluster_limit;				//when the number of residues in one residue cluster is greater than max_cluster_limit, use mask to deal with this big cluster.
	float	*wrapped_phase, *qual;			//interferogram, coherence coefficient matrix

	short	**unwrapped_k;                  //solution of phase unwrapping, 2-D matrix
	unsigned char *bitflags;			    	//matrix bitflags is used to mark residues and need_to_do, etc.
	int		*mark_global;					//matrix mark_global is used to store cluster information

	FILE	*pfp, *qfp, *ofp;													//file pointer
	char	phase_file[MAXSTRLEN], qual_file[MAXSTRLEN], out_file[MAXSTRLEN];   	//file path
	char	conf_file[] = "yu.cfg";												//configuration file name, which has a fixed name "yu.cfg"
	int		npara;      														//Parameter statistics read from the configuration file
	int     res_num; 										     				//Global total number of residues

	short	*row,*col;						//
	int		*type;							//
	float	*dist;							//
	clock_t btime,endtime;					// running time statistics
	int		bad_flag = FALSE;				// bad_flag is used to mark whether there is a large bad area, which affects the integral function

	fprintf(STDOUT, ODLS);
	/*********************************  Read Configuration File **************************************************************/
	fprintf(STDOUT, "Reading configuration file...\n");
	npara = ReadConfigFile(conf_file, phase_file, qual_file, out_file, &xsize, &lowdis, &highdis, &delta, &max_cluster_limit);
	fprintf(STDOUT, "%d parameters read from configuration file\n",npara);
	fprintf(STDOUT, "Wrapped phase file: \t%s\n", phase_file);
	if (strlen(qual_file)) {
		fprintf(STDOUT, "Quality file: \t\t%s\n", qual_file);
	}
	fprintf(STDOUT, "Output file: \t\t%s\n", out_file);
	fprintf(STDOUT, "High/Low initial cluster distance: %.2f(L), %.2f(H)\n", lowdis, highdis);
	fprintf(STDOUT, "Cluster distance step: \t\t%.2f\n", delta);
	fprintf(STDOUT, "Max cluster limit: \t\t%d\n", max_cluster_limit);

	/********************************* Read Wrapped Phase ********************************************************************/
	fprintf(STDOUT, "Reading wrapped phase...");
	OpenFile(&pfp, phase_file, "rb");
	ysize = GetYsize(pfp, xsize, sizeof(float));
	rewind(pfp);
	AllocateMemory(&wrapped_phase, xsize*ysize, sizeof(float));
	ReadData(pfp, wrapped_phase, xsize*ysize, sizeof(float));
	fprintf(STDOUT, " Done.\n");
	fprintf(STDOUT, "Data size: %d X %d\n", xsize, ysize);

	/****************************** Mark Residues in Global ******************************************************************/
	AllocateMemory(&bitflags, xsize*ysize, sizeof(unsigned char));
	memset(bitflags, 0, xsize*ysize*sizeof(unsigned char));
	fprintf(STDOUT, "Mark residues in bitflags...\n");
	res_num = Residues(wrapped_phase, bitflags, POS_RES, NEG_RES, AVOID, xsize, ysize);
	fprintf(STDOUT, "\n%d residues found.\n", res_num);
	free(wrapped_phase);

	/****************************** Read or Generate Coherence Coefficients *************************************************/
	AllocateMemory(&qual, xsize*ysize, sizeof(float));
	memset(qual, 0, xsize*ysize*sizeof(float));
	if (strlen(qual_file)) {
		fprintf(STDOUT, "Reading correlation...");
		OpenFile(&qfp, qual_file, "rb");
		ReadData(qfp, qual, xsize*ysize, sizeof(float));
		fprintf(STDOUT, " Done.\n");
	} else {
		fprintf(STDOUT, "No Correlation file, generating correlation...\n");
		GetQualMap(wrapped_phase, qual, xsize, ysize);
		OpenFile(&qfp, "corr.corr.in", "wb");				//
		WriteData(qfp, qual, xsize*ysize, sizeof(float));
		fprintf(STDOUT, "\nDone.\n");
	}

	/********************************* Start Clustering *********************************************************************/
	fprintf(STDOUT, "Clustering...\n");
	btime = clock();
#if ONECLUSTER
	AllocateMemory(&mark_global, xsize*ysize, sizeof(int));
 	memset(mark_global, 0, xsize*ysize*sizeof(int));
	Cluster(mark_global, bitflags, qual, xsize, ysize, res_num, delta, lowdis, highdis, &row, &col, &type, &dist);
#else
	ClusterRepeatErgodic(&mark_global, bitflags, qual, xsize, ysize, res_num, delta, lowdis, highdis, &row, &col, &type, &dist);
#endif
	endtime = clock();
	fprintf(STDOUT, "Done. Use %ld s\n", (endtime-btime)/CLOCKS_PER_SEC);

#if OTHERWORK
	
	/***************************** Generate a MST for Each cluster **********************************************************/
	// MST, minimmum spanning tree
	bad_flag = GenMSTforAllCluster(mark_global, bitflags, row, col, type, dist, res_num, xsize, ysize, max_cluster_limit);
	free(row);
	free(col);
	free(type);
	free(dist);
#if BIGKIND
	FindLargeKind(bitflags, mark_global, xsize, ysize, res_num);
#endif
	free(mark_global);

	// To save memory, reload the data of wrapped phase
	OpenFile(&pfp, phase_file, "rb");
	AllocateMemory(&wrapped_phase, xsize*ysize, sizeof(float));
	ReadData(pfp, wrapped_phase, xsize*ysize, sizeof(float));
	Allocate2DArray(&unwrapped_k, xsize, ysize, sizeof(short));

	/********************************* Global Integration *******************************************************************/
	// The integration path does not cross through any branch cut generated by MST.
#if BIGKIND
	SolveLargeKind(wrapped_phase, unwrapped_k, bitflags, xsize, ysize);
#else
	PathIntegAroundCuts(wrapped_phase, unwrapped_k, bitflags, xsize, ysize, bad_flag);
#endif
	Free2DArray(unwrapped_k, xsize);
#else
	free(row);
	free(col);
	free(type);
	free(dist);
	free(mark_global);
#endif
	free(bitflags);
	
#if OTHERWORK
	/****************************** Save the PU solution ********************************************************************/
	fprintf(STDOUT, "Saving unwrapped phase...\n");
	OpenFile(&ofp, out_file, "wb");
	WriteData(ofp, wrapped_phase, xsize*ysize, sizeof(float));
	free(wrapped_phase);
	
	endtime = clock();
	fprintf(STDOUT, "Done.Total consume %ld s\n", (endtime-btime)/CLOCKS_PER_SEC);
#endif
	
	fprintf(STDOUT, "Press <Enter> to exit.");
	getchar();
	exit(EXIT_SUCCESS);
}