/*************************************************************************************************************************/
/* Description:
/* This program is used to solve the kind block (i.e., high quality area) and clusters by integrating.

/* Release date of FLSv1.0: Aug.05, 2018
/* Latest updated date of FLSv2.0: Oct.02, 2018
/*************************************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "file.h"
#include "solver.h"

// solve the largest good area by integrating
void SolveKindBlock(float *phase, float *unwrapped, int *mark_global, unsigned char *bitflags, 
					const int xsize, const int ysize, const int resnum, int kind_num) {

	int i, j, k, ii, jj, kk, ka, boxi, boxj;
	int *active_list, active_num=0;
	float delta;

	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			if (mark_global[i*ysize+j]==kind_num) {
				bitflags[i*ysize+j] |= NEED_TODO;
			} else {
				bitflags[i*ysize+j] &= ~NEED_TODO;
			}
		}
	}

	fprintf(STDOUT, "Solving biggest kind block...\n");
	AllocateMemory(&active_list, xsize*ysize-resnum, sizeof(int));
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if ((bitflags[k]&NEED_TODO)) {
				active_num = 0;
				active_list[active_num++] = k;
				bitflags[k] &= ~NEED_TODO;
				bitflags[k] |= CANBE_INIT;
				mark_global[k] = 0;
				unwrapped[k] = phase[k];

				for (ka = 0; ka < active_num; ka++) {

					boxi = active_list[ka]/ysize;
					boxj = active_list[ka]%ysize;
					for (ii = boxi-1; ii <= boxi+1; ii++) {
						for (jj = boxj-1; jj <= boxj+1; jj++) {
							if (ii<0 || ii>=xsize || jj<0 || jj>=ysize) {	//
								continue;
							}
							if (ii==boxi || jj==boxj) {						//
								kk = ii*ysize+jj;
								if (bitflags[kk]&NEED_TODO) {
									delta = phase[kk]-phase[active_list[ka]];
									if (delta >= PI) {
										delta -= (float)(2*PI);
									} else if (delta <= -PI) {
										delta += (float)(2*PI);
									}
									active_list[active_num++] = kk;
									bitflags[kk] &= ~NEED_TODO;
									bitflags[kk] |= CANBE_INIT;
									mark_global[kk] = 0;
									unwrapped[kk] = unwrapped[active_list[ka]]+delta;
								}
							}
						}
					}
				} //end of ka...
				goto over;
			} //end of if...
		} //end of j...
	} //end of i...
over:
	free(active_list);
	fprintf(STDOUT, "Done.\n");
}

// count all the residues that need to be solved
int NumOfNeedToDo(unsigned char *bitflags, const int xsize, const int ysize) {

	int i, j, count=0;

	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			if (bitflags[i*ysize+j]&NEED_TODO) {
				++count;
			}
		}
	}
	return count;
}

// generate gradient matrix A and gradient vector b
void GenerageCoeff(float *phase, unsigned char *bitflags, const int xsize, const int ysize, 
				   double **row, double **col, double **val, double **b, 
				   int *len_a, int *len_b, int *len_x, int *len_y) {
	int *pos_vec, vec_size;
	int i, j, k, ka;
	int temp[5];
	double temp_b;
	LinkedQueue *Qa;
	LinkedNode *node;
	int count_a;

	Qa = (LinkedQueue *)malloc(sizeof(LinkedQueue));
	InitQueue(Qa);
	
	vec_size = NumOfNeedToDo(bitflags, xsize, ysize);
	*len_b = vec_size;
	AllocateMemory(&pos_vec, vec_size, sizeof(int));
	
	ka = 0;
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if (bitflags[k] & NEED_TODO) {
				pos_vec[ka++] = k;
			}
		}
	}
	
	ka = 1;
	count_a = 0;
	for (i = 0; i < xsize; i++) {
		for (j = 0; j < ysize; j++) {
			k = i*ysize+j;
			if (bitflags[k] & NEED_TODO) {
				if (bitflags[k] & CANBE_INIT) {
					temp[0] = ka++;
					temp[1] = BinSearch(pos_vec, 0, vec_size, k)+1;
					temp[2] = -1;
					temp[3] = k;
					temp[4] = 0;
					EnQueue(Qa, temp, 5);
					++count_a;
				}
				if ((j+1)<ysize && bitflags[k+1]&NEED_TODO) {
					temp[0] = ka++;
					temp[1] = BinSearch(pos_vec, 0, vec_size, k)+1;
					temp[2] = BinSearch(pos_vec, 0, vec_size, k+1)+1;
					temp[3] = k;
					temp[4] = k+1;
					EnQueue(Qa, temp, 3);
					count_a += 2;
				}
				if ((i+1)<xsize && bitflags[k+ysize]&NEED_TODO) {
					temp[0] = ka++;
					temp[1] = BinSearch(pos_vec, 0, vec_size, k)+1;
					temp[2] = BinSearch(pos_vec, 0, vec_size, k+ysize)+1;
					temp[3] = k;
					temp[4] = k+ysize;
					EnQueue(Qa, temp, 3);
					count_a += 2;
				}
			}
		}
	}
	
	AllocateMemory(row, count_a, sizeof(double));
	AllocateMemory(col, count_a, sizeof(double));
	AllocateMemory(val, count_a, sizeof(double));
	AllocateMemory(b, ka, sizeof(double));

	*len_x = ka;
	*len_y = vec_size;

	node = Qa->head;
	ka = 0;
	count_a = 0;
	while (node != NULL) {
		if (node->elem[2]==-1) {
			(*row)[count_a] = (double)node->elem[0];
			(*col)[count_a] = (double)node->elem[1];
			(*val)[count_a] = 1.0;
			++count_a;
			(*b)[ka++] = phase[node->elem[3]];
			node = node->next;
		} else {
			(*row)[count_a] = (double)node->elem[0];
			(*col)[count_a] = (double)node->elem[1];
			(*val)[count_a] = 1.0;
			++count_a;
			(*row)[count_a] = (double)node->elem[0];
			(*col)[count_a] = (double)node->elem[2];
			(*val)[count_a] = -1.0;

			temp_b = phase[node->elem[3]]-phase[node->elem[4]];
			if (temp_b >= PI) {
				temp_b -= 2*PI;
			} else if (temp_b <= -PI) {
				temp_b += 2*PI;
			}
			(*b)[ka++] = temp_b;
			++count_a;
		}
	}
	DestroyQueue(Qa);
	free(Qa);
	free(pos_vec);
}
