/*************************************************************************************************************************/
/* Description:
/* These are some general functions.

/* Release date of FLSv1.0: Aug.05, 2018
/* Latest updated date of FLSv2.0: Oct.02, 2018
/*************************************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <ctype.h>
#include "util.h"
#include "file.h"

// function to handle error messages
void ErrorHandle(char message[], int exit_code) {

	fprintf(STDERR, "\n%s.\nAbort\n", message);
	getchar();
	exit(exit_code);
}

// function for opening a file
void OpenFile(FILE **fp, const char *filepath, const char *mode) {

	// *fp = fopen(filepath, mode);
	fopen_s(fp,filepath, mode);
	if (*fp == NULL) {
		ErrorHandle("Cannot open file", OPEN_FILE_ERROR);
	}
}

// function for allocating 1-D memory
void AllocateMemory(void **array, const int len, const int elemsize) {

	*array = malloc(len*elemsize);
	if (*array == NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
}

// function for releasing 2-D memory
void Free2DArray(void **array, const int xsize) {

	int i;
	for (i = 0; i < xsize; ++i) {
		free(array[i]);
	}
	free(array);
}

// function for allocating 2-D memory
void Allocate2DArray(void ***array, const int xsize, const int ysize, const int elemsize) {

	int i;
	*array = (void **)malloc(xsize*sizeof(void *));
	if (*array == NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	for (i = 0; i < xsize; ++i) {
		(*array)[i] = (void *)malloc(ysize*elemsize);
		if ((*array)[i] == NULL) {
			ErrorHandle("Out of memory", OUT_OF_MEMORY);
		}
		memset((*array)[i], 0, ysize*elemsize);
	}
}

// function for writing data in 2-D memory to a file
void Write2DArray(FILE *fp, void **array, const int xsize, const int ysize, const int elemsize) {

	int i;
	for (i = 0; i < xsize; ++i) {
		if (ysize != fwrite(array[i], elemsize, ysize, fp)) {
			ErrorHandle("Write data fail", WRITE_ERROR);
		}
	}
	fclose(fp);
}

// function for reading data into 1-D memory
void ReadData(FILE *fp, void *data, const int len, const int elemsize) {

	if (len != fread(data, elemsize, len, fp)) {
		ErrorHandle("Read data fail", READ_ERROR);
	}
	fclose(fp);
}

// function for writing data in 1-D memory to a file
void WriteData(FILE *fp, void *data, const int len, const int elemsize) {

	if (len != fwrite(data, elemsize, len, fp)) {
		ErrorHandle("Write data fail", WRITE_ERROR);
	}
	fclose(fp);
}

// function for correcting the gradient
float Gradient(float num1, float num2) {
	
	float result;
	result = num1 - num2;
	if (result > PI) {
		result -= 2*PI;
	} else if (result < -PI) {
		result += 2*PI;
	}
	return result;
}

// function for reading the configuration file
int ReadConfigFile(const char *conf_file, char *phase_file, char *qual_file, char *out_file, int *xsize, 
				   float *lowdis, float *highdis, float *delta, int *cluster_limit) {

	FILE *cfp;
	int necessary = 0, nfields, paranum = 0;
	char buf[MAXSTRLEN], para1[MAXSTRLEN], para2[MAXSTRLEN];
	char *line_str;
	long ltemp;
	double dtemp;
	int qual_file_flag = FALSE;

	OpenFile(&cfp, conf_file, "r");
	
	while (TRUE) {
		buf[0] = '\0';
		line_str = fgets(buf, MAXSTRLEN, cfp);
		if(line_str == NULL || (!strlen(line_str))) {
			break;
		}
		
		nfields = sscanf(buf, "%s %s",para1,para2);
		if (nfields == 1 && isalnum(para1[0])) {
			ErrorHandle("Wrong config information", CONFIG_ERROR);
		}
		
		if (nfields == 2 && isalnum(para1[0])) {
			++paranum;
			if (!strcmp(para1, "PHASEFILE")) {
				strncpy(phase_file, para2, MAXSTRLEN);
				++necessary;
			} else if (!strcmp(para1, "QUALFILE")) {
				strncpy(qual_file, para2, MAXSTRLEN);
				qual_file_flag = TRUE;
			} else if (!strcmp(para1, "OUTFILE")) {
				strncpy(out_file, para2, MAXSTRLEN);
				++necessary;
			} else if (!strcmp(para1, "XSIZE")) {
				if (StrToLong(para2, &ltemp)) {
					*xsize = ltemp;
					++necessary;
				} else {
					ErrorHandle("Convert to long error", RUNTIME_ERROR);
				}
			} else if (!strcmp(para1, "LOWDIS")) {
				if (StrToDouble(para2, &dtemp)) {
					*lowdis = dtemp;
					++necessary;
				} else {
					ErrorHandle("Convert to double error", RUNTIME_ERROR);
				}
			} else if (!strcmp(para1, "HIGHDIS")) {
				if (StrToDouble(para2, &dtemp)) {
					*highdis = dtemp;
					++necessary;
				} else {
					ErrorHandle("Convert to double error", RUNTIME_ERROR);
				}
			} else if (!strcmp(para1, "DELTA")) {
				if (StrToDouble(para2, &dtemp)) {
					*delta = dtemp;
					++necessary;
				} else {
					ErrorHandle("Convert to double error", RUNTIME_ERROR);
				}
			} else if (!strcmp(para1, "MAXCLUSTER")) {
				if (StrToLong(para2, &ltemp)) {
					*cluster_limit = ltemp;
					++necessary;
				} else {
					ErrorHandle("Convert to long error", RUNTIME_ERROR);
				}
			} else {
				fprintf(STDOUT, "Warning: unknown parameter \'%s\'\n", para1);
			}
		}
	}
	
	if (necessary < 7) {
		fprintf(STDERR, "Too few parameters in configuration file.\nAbort.\n");
		exit(CONFIG_ERROR);
	}

	if (!qual_file_flag) {
		strncpy(qual_file, "", MAXSTRLEN);
	}
	return paranum;
}

// data type conversion: string to Double
int StrToDouble(char *str, double *dnum) {
	double dtmp;
	char *endp;
	endp = str;
	dtmp = strtod(str, &endp);
	if (strlen(endp) || dtmp == DBL_MAX || dtmp == DBL_MIN) {
		return FALSE;
	}
	*dnum = dtmp;
	return TRUE;
}

// data type conversion: string to long integer
int StrToLong(char *str, long *lnum) {
	long ltmp;
	char *endp;
	endp = str;
	ltmp = strtol(str, &endp, 10);
	if (strlen(endp) || ltmp == LONG_MAX || ltmp == LONG_MIN) {
		return FALSE;
	}
	*lnum = ltmp;
	return TRUE;
}

// 
int GetYsize(FILE *pfp, int xsize, int elemsize) {

	int ysize;
	long filesize;
	fseek(pfp, 0, SEEK_END);
	filesize = ftell(pfp);
	rewind(pfp);
	ysize = filesize / elemsize / xsize;
	return ysize;
}


//  Linear Sort
void LinearSortNew(short *row, short *col, int *type, float *dist, const int len) {
	
	int i, j;
	int k, output[3];
	int *C;
	short *outrow, *outcol;
	int *outtype;
	float *outdist;
	
	Maximum(type,1,len,output);
	k = output[2];
	AllocateMemory(&C, k+1, sizeof(int));
	
    for(i=0; i<=k; i++){
		C[i] = 0;
	}
	for(j=0; j<len; j++){
		C[type[j]] = C[type[j]]+1;
	}
	//now C[i] contains the num of elements equals to i
	for(i=1; i<=k; i++){
		C[i] = C[i]+C[i-1];
	}
	//now C[i] contains the num of <= to i
	
	AllocateMemory(&outrow, len, sizeof(short));
	AllocateMemory(&outcol, len, sizeof(short));
	AllocateMemory(&outtype, len, sizeof(int));
	AllocateMemory(&outdist, len, sizeof(float));
	for(j=len-1; j>=0; j--){
		outrow[C[type[j]]-1]  = row[j];
		outcol[C[type[j]]-1]  = col[j];
		outdist[C[type[j]]-1] = dist[j];
		outtype[C[type[j]]-1] = type[j];
		
		C[type[j]] = C[type[j]] - 1;
	}
	
	memcpy(row, outrow, len*sizeof(short));
	memcpy(col, outcol, len*sizeof(short));
	memcpy(type, outtype, len*sizeof(int));
	memcpy(dist, outdist, len*sizeof(float));
	free(C);
	free(outrow);
	free(outcol);
	free(outtype);
	free(outdist);
}

void Maximum(int *array, int xsize, int ysize, int *output)
{
	int i=0,j=0,k=0;
	int max_x_index = 0, max_y_index = 0;
	int mm = array[0];
	
	// find max
	for(k=0; k<(xsize*ysize); k++){
		if(array[k] > mm){
			mm = array[k];
			max_x_index = (int)(k/ysize);
			max_y_index = (k%ysize);
		}
	}
	output[0] = max_x_index;
	output[1] = max_y_index;
	output[2] = mm;
}

// binary search
int BinSearch(int *array, int low, int high, int elem) {

	int mid;
	if (low < 0 || high < 0) {
		ErrorHandle("Wrong position", RUNTIME_ERROR);
	}
	while (low <= high) {
		mid = (low+high)/2;
		if (array[mid] == elem) {
			return mid;
		} else if (array[mid] < elem) {
			low = mid+1;
		} else {
			high = mid-1;
		}
	}
	return -1;
}

// renumber the residue cluster and remove the number of empty cluster
void ReKind(int *type, const int len) {
	
	int i, cur_type = type[0], max_type = 1;
	
	for (i = 0; i < len; ++i) {
		if (type[i] == cur_type) {
			type[i] = max_type;
		} else {
			cur_type = type[i];
			type[i] = ++max_type;
		}
	}
}

//
void InitQueue(LinkedQueue *Q) {

	if (Q == NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	Q->head = Q->tail = (LinkedNode *)malloc(sizeof(LinkedNode));
	if (Q->head == NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	Q->head->next = NULL;
}

//
void DestroyQueue(LinkedQueue *Q) {

	while (Q->head->next != NULL) {
		Q->tail = Q->head->next;
		free(Q->head);
		Q->head = Q->tail;
	}
}

// insert elements in the queue
void EnQueue(LinkedQueue *Q, int *new_data, int dim) {

	LinkedNode *node = (LinkedNode *)malloc(sizeof(LinkedNode));
	if (node==NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	node->elem = (int *)malloc(dim*sizeof(int));
	if (node->elem==NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	memcpy(node->elem, new_data, dim*sizeof(int));
	node->dim = dim;
	node->next = NULL;
	Q->tail->next = node;
	Q->tail = node;
}

// the following queue is dedicated queue designed for PU
void InitIntQueue(IntQueue *Q) {

	if (Q == NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	Q->head = Q->tail = (IntNode *)malloc(sizeof(IntNode));
	if (Q->head == NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	Q->head->next = NULL;
}

void DestroyIntQueue(IntQueue *Q) {

	while (Q->head->next != NULL) {
		Q->tail = Q->head->next;
		free(Q->head);
		Q->head = Q->tail;
	}
}

void EnIntQueue(IntQueue *Q, int data) {

	IntNode *node = (IntNode *)malloc(sizeof(IntNode));
	if (node==NULL) {
		ErrorHandle("Out of memory", OUT_OF_MEMORY);
	}
	node->data = data;
	node->next = NULL;
	Q->tail->next = node;
	Q->tail = node;
}

int DeIntQueue(IntQueue *Q) {

	IntNode *p;
	int e;

	if (Q->head == Q->tail) {
		return -1;
	}
	p = Q->head->next;
	e = p->data;
	Q->head->next = p->next;
	if (Q->tail == p) {
		Q->tail = Q->head;
	}
	free(p);
	return e;
}